"""Provides the Artifact class."""
from typing import TYPE_CHECKING, Optional, Dict, Any, Union

from .base import DriveBase, Downloadable
from ..base import BaseList
from ...routing import Route, ENDPOINTS

if TYPE_CHECKING:
    from .. import TekDrive


class Artifact(DriveBase, Downloadable):
    """
    Represents a file artifact.

    Attributes:
        bytes (str): Artifact size in bytes.
        created_at (datetime): When the artifact was created.
        context_type (str): Context to identify the type of artifact such as ``"SETTING"`` or ``"CHANNEL"``.
        file_id (str): ID of the file that the artifact is associated with.
        file_type (str): Artifact file type such as ``"TSS"`` or ``"SET"``.
        id (str): Unique artifact ID.
        name (str): Artifact name.
        parent_artifact_id: ID of the parent artifact.
        updated_at (datetime, optional): When the artifact was last updated.
    """

    STR_FIELD = "id"

    @classmethod
    def from_data(cls, tekdrive, data):
        return cls(tekdrive, data)

    def __init__(
        self,
        tekdrive: "TekDrive",
        _data: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(tekdrive, _data=_data)

    def __setattr__(
        self,
        attribute: str,
        value: Union[str, int, Dict[str, Any]],
    ):
        super().__setattr__(attribute, value)

    def _fetch_download_url(self):
        route = Route("GET", ENDPOINTS["file_artifact_download"], file_id=self.file_id, artifact_id=self.id)
        download_details = self._tekdrive.request(route)
        return download_details["download_url"]


class ArtifactsList(BaseList):
    """List of artifacts"""

    _parent = None

    LIST_ATTRIBUTE = "artifacts"



"""Provides the Plan class."""
from typing import TYPE_CHECKING, Optional, Dict, Any

from .base import DriveBase
from ...enums import SharingType

if TYPE_CHECKING:
    from .. import TekDrive


class Plan(DriveBase):
    """
    Represents a TekDrive plan.

    Attributes:
        id (str): Unique ID of the plan.
        name (str): Plan name.
        limits (dict): Raw limits for the plan such as storage limit,
            access key limit, and sharing type.
    """

    STR_FIELD = "id"

    @classmethod
    def from_data(cls, tekdrive, data):
        return cls(tekdrive, data)

    def __init__(
        self,
        tekdrive: "TekDrive",
        _data: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(tekdrive, _data=_data)

    @property
    def access_key_limit(self) -> int:
        """
        The number of access keys allowed by the plan.
        """
        return self.limits.get("access_key_limit")

    @property
    def sharing_type(self) -> SharingType:
        """
        The plan sharing type.
        """
        return SharingType(self.limits["sharing_type"])

    @property
    def storage_limit(self) -> int:
        """
        The plan storage limit in bytes.
        """
        return int(self.limits["storage_size_limit"])



"""Provides the Folder class."""
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, IO, Optional, List, Union

from ...routing import Route, ENDPOINTS
from ...utils.casing import to_snake_case, to_camel_case
from .base import DriveBase
from ...exceptions import ClientException
from ...enums import FolderType, ObjectType
from .member import Member
from ..permissions import Permissions
from .user import PartialUser
from .file import File

if TYPE_CHECKING:
    from .. import TekDrive


class Folder(DriveBase):
    """
    A class representing a TekDrive folder.

    Examples:
        Load a folder by id::

            folder_id = "3b525331-9da7-4e8d-b045-4acdba8d9dc7"
            folder = td.folder(folder_id)

    Attributes:
        created_at (datetime): When the folder was created.
        creator (:ref:`partial_user`): Folder creator.
        id (str): Unique ID of the folder.
        folder_type (str): Folder type such as ``"STANDARD"``, ``"PERSONAL"``, or ``"SHARES"``.
        name (str): Name of the folder.
        owner (:ref:`partial_user`): Folder owner.
        parent_folder_id (str): Unique ID of the folder's parent folder.
        shared_at (datetime, optional): When the folder was shared with the
            requesting user. Will be ``None`` if the user has direct access.
        type (str): Type of TekDrive object - will always be ``"FOLDER"``.
        permissions (:ref:`permissions`): Folder permissions for the requesting user.
        updated_at (datetime, optional): When the folder was last updated.
    """

    STR_FIELD = "id"

    def __init__(
        self,
        tekdrive: "TekDrive",
        id: Optional[str] = None,
        _data: Optional[Dict[str, Any]] = None,
    ):
        fetched = False
        if id:
            self.id = id
        else:
            fetched = True
        self._children = None
        super().__init__(tekdrive, _data=_data, _fetched=fetched)

    def __setattr__(
        self,
        attribute: str,
        value: Union[str, int, "Member"],
    ):
        if attribute == "owner" or attribute == "creator":
            value = PartialUser(**value)
        elif attribute in ("created_at", "updated_at", "shared_at"):
            if value is not None:
                value = datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ")
        elif attribute == "permissions":
            value = Permissions(**value)
        elif attribute == "type":
            value = ObjectType(value)
        elif attribute == "folder_type":
            value = FolderType(value)
        elif attribute == "children":
            self._children = [
                File(self._tekdrive, d["id"], _data=d)
                if d.get("type") == "FILE"
                else Folder(self._tekdrive, d["id"], _data=d)
                for d in value
            ]
            return
        super().__setattr__(attribute, value)

    def _fetch_data(self):
        route = Route("GET", ENDPOINTS["folder_details"], folder_id=self.id)
        return self._tekdrive.request(route, should_parse=False)

    def _fetch(self):
        data = self._fetch_data()
        other = type(self)(self._tekdrive, _data=to_snake_case(data))
        self.__dict__.update(other.__dict__)
        self._fetched = True

    def _update_details(self, data):
        route = Route("PUT", ENDPOINTS["folder_details"], folder_id=self.id)
        return self._tekdrive.request(route, json=data)

    @staticmethod
    def _create(
        _tekdrive,
        name: str = None,
        parent_folder_id: str = None,
    ) -> "Folder":
        data = dict(name=name, parentFolderId=parent_folder_id)
        route = Route("POST", ENDPOINTS["folder_create"])
        new_folder = _tekdrive.request(route, json=data)
        return new_folder

    def children(self) -> List[Union[File, "Folder"]]:
        """
        Get a list of child files and folders for the given folder.

        Examples:
            Iterate over all folder children::

                for child in folder.children():
                    print(child.id)
        
        Returns:
            List [ Union [:ref:`file` , :ref:`folder`] ]
        """
        if self._children is not None:
            return self._children

        params = to_camel_case({"folder_id": self.id})
        route = Route("GET", ENDPOINTS["tree"])
        return self._tekdrive.request(route, params=params)._children

    def members(self) -> List[Member]:
        """
        Get a list of folder members.

        Examples:
            Iterate over all folder members::

                for member in folder.members():
                    print(member.username)

        Returns:
            List [ :ref:`member` ]
        """
        route = Route("GET", ENDPOINTS["folder_members"], folder_id=self.id)
        members = self._tekdrive.request(route)
        members._parent = self
        return members

    def restore(self) -> None:
        """
        Restore the folder from user's trashcan.

        Examples:
            Restore - remove from trash::

                folder.restore()
        """
        route = Route("POST", ENDPOINTS["folder_restore"], folder_id=self.id)
        self._tekdrive.request(route)

    def delete(self, hard_delete: bool = False) -> None:
        """
        Delete the folder, by default it will be placed in the user's trashcan.

        Args:
            hard_delete: Permanently delete the folder?

        Examples:
            Delete - placing in trash::

                folder.delete()

            Delete - immediately delete::

                folder.delete(hard_delete=True)
        """
        params = to_camel_case(dict(hard_delete=hard_delete))

        route = Route("DELETE", ENDPOINTS["folder_delete"], folder_id=self.id)
        self._tekdrive.request(route, params=params)

    def move(self, parent_folder_id: str) -> None:
        """
        Move folder.

        Args:
            parent_folder_id: Unique ID of another folder to move the folder into.

        Examples:
            Move by ID::

                new_parent_folder_id = "1ae7fa38-9759-49f1-9e41-28a190b96023"
                folder.move(new_parent_folder_id)

            Move to newly created folder::

                new_parent_folder = td.folder.create("FolderA")
                folder.move(new_parent_folder.id)
        """
        data = dict(parentFolderId=parent_folder_id)
        self._update_details(data)
        self.parent_folder_id = parent_folder_id

    def save(self) -> None:
        """
        Save any changes to folder meta. Supported attributes: ``name``.

        Examples:
            Rename a folder::

                folder.name = "my_new_name"
                folder.save()
        """
        data = dict(name=self.name)
        self._update_details(data)

    def add_member(
        self, username: str = None, user_id: str = None, edit_access: bool = False
    ) -> Member:
        """
        Share the folder with an existing or new user.

        Args:
            username: The username (email) of the sharee.
            user_id: The user ID of the sharee.
            edit_access: Give member edit access?

        Examples:
            Share with read only permissions (default)::

                folder_member = folder.add_member(username="read_only@example.com")

            Share with edit permissions::

                folder_member = folder.add_member(user_id="354bcafb-6c54-4a1f-9b94-a76f38b548e5", edit_access=True)

        Returns:
            :ref:`member`
        """
        data = dict(permissions=dict(read=True, edit=edit_access))
        if user_id:
            data["id"] = user_id
        elif username:
            data["username"] = username
        else:
            raise ClientException("Must supply `username` or `user_id`.")

        route = Route("POST", ENDPOINTS["folder_members"], folder_id=self.id)
        return self._tekdrive.request(route, json=data)

    def remove_member(self, user_id: str) -> None:
        """
        Revoke access for a current folder member.

        Args:
            user_id: The user ID of the member.

        Examples:
            Remove member::

                folder.remove_member(user_id="354bcafb-6c54-4a1f-9b94-a76f38b548e5")

        """
        route = Route(
            "DELETE", ENDPOINTS["folder_member"], folder_id=self.id, member_id=user_id
        )
        self._tekdrive.request(route)

    def modify_member(self, user_id: str, edit_access: bool) -> Member:
        """
        Modify an existing folder member.

        Args:
            user_id: The user ID of the member.
            edit_access: Give member edit access?

        Examples:
            Grant edit access::

                updated_folder_member = folder.modify_member(user_id="354bcafb-6c54-4a1f-9b94-a76f38b548e5", edit_access=True)

            Revoke edit access::

                updated_folder_member = folder.modify_member(user_id="354bcafb-6c54-4a1f-9b94-a76f38b548e5", edit_access=False)

        Returns: 
            :ref:`member`
        """
        route = Route(
            "PUT", ENDPOINTS["folder_member"], folder_id=self.id, member_id=user_id
        )
        data = dict(permissions=dict(read=True, edit=edit_access))
        return self._tekdrive.request(route, json=data)

    def upload(self, path_or_readable: Union[str, IO], file_name: str):
        """
        Create a new file and automatically upload its contents directly into this folder.

        Args:
            path_or_readable: Path to a local file or a readable stream
                representing the contents to upload.
            file_name (optional): Name of the new file.

        Raises:
            ClientException: If invalid file path is given.

        Examples:
            Upload using path::

                here = os.path.dirname(__file__)
                contents_path = os.path.join(here, "test_file_overwrite.txt")
                folder.upload(contents_path, "my_new_file")

            Upload using readable stream::

                with open("./test_file.txt"), "rb") as f:
                    folder.upload(f)
        """
        return self._tekdrive.file.create(
            path_or_readable, name=file_name, parent_folder_id=self.id
        )



from enum import Enum


class SharingType(Enum):
    UNLIMITED = "UNLIMITED"
    PERMISSIONS_LIMITED = "PERMISSIONS_LIMITED"
    DISABLED = "DISABLED"


class ObjectType(Enum):
    ARTIFACT = "ARTIFACT"
    FILE = "FILE"
    FOLDER = "FOLDER"


class FolderType(Enum):
    STANDARD = "STANDARD"
    SHARES = "SHARES"
    PERSONAL = "PERSONAL"


class ErrorCode(Enum):
    ARTIFACT_NOT_FOUND = "ARTIFACT_NOT_FOUND"
    FILE_GONE = "FILE_GONE"
    FILE_NOT_FOUND = "FILE_NOT_FOUND"
    FOLDER_GONE = "FOLDER_GONE"
    FOLDER_NOT_FOUND = "FOLDER_NOT_FOUND"
    FORBIDDEN = "FORBIDDEN"
    UNPROCESSABLE_ENTITY = "UNPROCESSABLE_ENTITY"



"""Provides the Tree class."""
from typing import TYPE_CHECKING, Optional

from ..routing import Route, ENDPOINTS
from .base import TekDriveBase
from . import Folder
from ..utils.casing import to_camel_case

if TYPE_CHECKING:
    from .. import TekDrive


class Tree(TekDriveBase):
    """
    Provides directory listing.
    """

    def __init__(self, tekdrive: "TekDrive"):
        super().__init__(tekdrive=tekdrive, _data=None)

    def get(
        self,
        *,
        folder_id: [str] = None,
        silo: Optional[str] = None,
        depth: Optional[int] = 1,
        folders_only: bool = False,
        include_trashed: bool = False,
    ) -> Folder:
        """
        Get the tree representation from a starting folder.

        Args:
            folder_id: Unique ID of starting folder to start the tree from.
            depth: How many nested levels to return.
            silo: Get tree for the provided silo. Values: ``"SHARES"`` or ``"PERSONAL"``.
            folders_only: Only include folders in the tree results? Default: ``False``.
            include_trashed: Include files and folders that are in the trashcan.

        Examples:
            Get tree from starting folder by id::

                tree = td.tree.get(folder_id="3b525331-9da7-4e8d-b045-4acdba8d9dc7")

            Get tree for the ``SHARES`` silo, excluding files::

                tree = td.tree.get(silo="SHARES", folders_only=True)

        Returns:
            :ref:`folder`
        """

        route = Route("GET", ENDPOINTS["tree"])
        params = to_camel_case(
            dict(
                folder_id=folder_id,
                silo=silo,
                depth=depth,
                folders_only=folders_only,
                include_trashed=include_trashed,
            )
        )
        return self._tekdrive.request(route, params=params)



"""Provide interface for HTTP request handling."""
import requests
from .settings import __version__
from .settings import TIMEOUT
from .exceptions import RequestException


class RequestWrapper(object):
    def __init__(
        self,
        base_url: str = "https://drive.api.tekcloud.com",
        session: requests.Session = None,
    ):
        """
        Args:
            base_url: The base URL for API requests. Default: https://drive.api.tekcloud.com.
            session: A session to handle requests
        """
        self._http = session or requests.Session()
        self._http.headers["User-Agent"] = f"pytekdrivecore/{__version__}"
        self._http.headers["Accept-Version"] = "v1"

        self.base_url = base_url

    def close(self):
        return self._http.close()

    def call(self, *args, timeout: float = TIMEOUT, **kwargs):
        """
        Make the HTTP request and capture errors, if any.
        """
        try:
            return self._http.request(*args, timeout=timeout, **kwargs)
        except Exception as exc:
            raise RequestException(exc, args, kwargs)



"""Provide the TekDriveBase model"""
from typing import TYPE_CHECKING, Any, Dict, Iterator, Optional

if TYPE_CHECKING:
    from .. import TekDrive


class TekDriveBase:
    @classmethod
    def parse(cls, data: Dict[str, Any], tekdrive: "TekDrive") -> Any:
        """
        Return an instance of ``cls`` from ``data``.

        Args:
            data: The structured data.
            tekdrive: An instance of :class:`.TekDrive`.

        """
        return cls(tekdrive, _data=data)

    def __init__(self, tekdrive: "TekDrive", _data: Optional[Dict[str, Any]]):
        self._tekdrive = tekdrive
        if _data:
            for attribute, value in _data.items():
                setattr(self, attribute, value)


class BaseList(TekDriveBase):

    LIST_ATTRIBUTE = None

    def __init__(self, tekdrive: "TekDrive", _data: Dict[str, Any]):
        super().__init__(tekdrive, _data=_data)

        if self.LIST_ATTRIBUTE is None:
            raise NotImplementedError("LIST_ATTRIBUTE is required.")

        child_list = getattr(self, self.LIST_ATTRIBUTE)
        for idx, item in enumerate(child_list):
            child_list[idx] = tekdrive._parser.parse(item)

    def __contains__(self, item: Any) -> bool:
        return item in getattr(self, self.LIST_ATTRIBUTE)

    def __getitem__(self, idx: int) -> Any:
        return getattr(self, self.LIST_ATTRIBUTE)[idx]

    def __iter__(self) -> Iterator[Any]:
        return getattr(self, self.LIST_ATTRIBUTE).__iter__()

    def __len__(self) -> int:
        return len(getattr(self, self.LIST_ATTRIBUTE))

    def __str__(self) -> str:
        return str(getattr(self, self.LIST_ATTRIBUTE))



"""Provide the DriveBase class."""
import os
import requests
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, IO, Optional, Union

from ..base import TekDriveBase
from ...exceptions import ClientException, TekDriveStorageException

if TYPE_CHECKING:
    from ... import TekDrive


class DriveBase(TekDriveBase):
    """Base class that represents actual TekDrive objects."""

    def __init__(
        self,
        tekdrive: "TekDrive",
        _data: Optional[Dict[str, Any]],
        _fetched: bool = False,
    ):
        """
        Initialize a DriveBase instance (or a subclass).

        Args:
            tekdrive: An instance of TekDrive.

        """
        super().__init__(tekdrive, _data=_data)
        self._fetched = _fetched
        if self.STR_FIELD not in self.__dict__:
            raise ValueError(f"An invalid value was specified for `{self.STR_FIELD}`.")

    def __eq__(self, other: Union[Any, str]) -> bool:
        """Return whether the other instance equals the current."""
        if isinstance(other, str):
            return other.lower() == str(self).lower()
        return (
            isinstance(other, self.__class__)
            and str(self).lower() == str(other).lower()
        )

    def __getattr__(self, attribute: str) -> Any:
        """Return the value of `attribute`."""
        if not attribute.startswith("_") and not self._fetched:
            self._fetch()
            return getattr(self, attribute)
        raise AttributeError(
            f"{self.__class__.__name__!r} object has no attribute {attribute!r}"
        )

    def __hash__(self) -> int:
        """Return the hash of the current instance."""
        return hash(self.__class__.__name__) ^ hash(str(self).lower())

    def __repr__(self) -> str:
        """Return an object initialization representation of the instance."""
        return f"{self.__class__.__name__}({self.STR_FIELD}={str(self)!r})"

    def __str__(self) -> str:
        """Return a string representation of the instance."""
        return getattr(self, self.STR_FIELD)

    def __ne__(self, other: Any) -> bool:
        """Return whether the other instance differs from the current."""
        return not self == other

    def _fetch(self):
        self._fetched = True

    def _reset_attributes(self, *attributes):
        for attribute in attributes:
            if attribute in self.__dict__:
                del self.__dict__[attribute]
        self._fetched = False


class Downloadable(ABC):
    """Abstract base class for download functionality."""

    @abstractmethod
    def _fetch_download_url(self):
        pass

    def _download_from_storage(self):
        download_url = self._fetch_download_url()
        try:
            r = requests.get(
                download_url,
            )
            r.raise_for_status()
            return r.content
        except requests.exceptions.HTTPError as exception:
            raise TekDriveStorageException("Download failed") from exception

    def download(self, path_or_writable: Union[str, IO] = None) -> None:
        """
        Download contents.

        Args:
            path_or_writable: Path to a local file or a writable stream
                where contents will be written.

        Raises:
            ClientException: If invalid file path is given.

        Examples:
            Download to local file using path::

                here = os.path.dirname(__file__)
                contents_path = os.path.join(here, "test_file_overwrite.txt")
                file.download(contents_path)

            Download using writable stream::

                with open("./download.csv", "wb") as f:
                    file.download(f)
        """
        if path_or_writable is None:
            # return content directly
            return self._download_from_storage()

        if isinstance(path_or_writable, str):
            file_path = path_or_writable

            if not os.path.exists(file_path):
                raise ClientException(f"File '{file_path}' does not exist.")

            with open(file_path, "wb") as file:
                file.write(self._download_from_storage())
        else:
            writable = path_or_writable
            writable.write(self._download_from_storage())



"""Provides the Trash class."""
from typing import TYPE_CHECKING, List, Optional

from ..routing import Route, ENDPOINTS
from .base import TekDriveBase
from .paginator import PaginatedListGenerator
from ..utils.casing import to_camel_case

if TYPE_CHECKING:
    from .. import TekDrive


class Trashcan(TekDriveBase):
    """
    Provides various methods to help with managing items in the trashcan.
    """

    def __init__(self, tekdrive: "TekDrive"):
        super().__init__(tekdrive=tekdrive, _data=None)

    def empty(
        self,
    ) -> None:
        """
        Empty all items currently in the trash.

        Examples:
            Empty the trashcan::

                td.trash.empty()
        """
        route = Route("DELETE", ENDPOINTS["trash"])
        self._tekdrive.request(route)

    def get(
        self,
        *,
        order_by: List[str] = ["-trashedAt"],
        limit: Optional[int] = 100,
    ):
        """
        Get items currently in the trash.

        Examples:
            Get the first 10 items in the trashcan::

                td.trash.get(limit=10)
        """
        route = Route("GET", ENDPOINTS["trash"])
        params = to_camel_case(
            dict(
                order_by=order_by,
            )
        )
        return PaginatedListGenerator(self._tekdrive, route, limit=limit, params=params)



"""Provide exception classes"""
from typing import TYPE_CHECKING

from .enums import ErrorCode

if TYPE_CHECKING:
    from requests import Response


class TekDriveException(Exception):
    """The base TekDrive Exception that all other exception classes extend."""


# API Exceptions
class TekDriveAPIException(TekDriveException):
    def __init__(
        self,
        data: dict,
        *,
        headers: dict = {},
    ):
        self.data = data
        self.headers = headers

    @property
    def error_code(self) -> str:
        return self.data.get("error_code")

    @property
    def message(self) -> str:
        return self.data.get("message")

    @property
    def errors(self):
        return self.data.get("errors")

    @property
    def request_id(self):
        return self.headers.get("X-Request-Id")


class ArtifactNotFoundAPIException(TekDriveAPIException):
    """Indicate artifact is not found."""


class FileNotFoundAPIException(TekDriveAPIException):
    """Indicate file is not found."""


class FileGoneAPIException(TekDriveAPIException):
    """Indicate file is gone."""


class ForbiddenAPIException(TekDriveAPIException):
    """Indicate action is forbidden."""


class FolderGoneAPIException(TekDriveAPIException):
    """Indicate file is gone."""


class FolderNotFoundAPIException(TekDriveAPIException):
    """Indicate folder is not found."""


class UnprocessableEntityAPIException(TekDriveAPIException):
    """Indicate server could not process the contained instructions."""


ERROR_CODE_TO_API_EXCEPTION_MAPPING = {
    ErrorCode.FILE_GONE: FileGoneAPIException,
    ErrorCode.FILE_NOT_FOUND: FileNotFoundAPIException,
    ErrorCode.FORBIDDEN: ForbiddenAPIException,
    ErrorCode.FOLDER_GONE: FolderGoneAPIException,
    ErrorCode.FOLDER_NOT_FOUND: FolderNotFoundAPIException,
    ErrorCode.UNPROCESSABLE_ENTITY: UnprocessableEntityAPIException,
}


class TekDriveStorageException(TekDriveException):
    """Indicate exceptions that happen uploading/downloading from storage."""


class ClientException(TekDriveException):
    """Indicate exceptions that happen client side."""


class InvalidAuthorizer(TekDriveException):
    """Indicate the authorizer is invalid and the request cannot be made."""


class HTTPException(TekDriveException):
    """Indicate an error with an HTTP request."""


class RequestException(HTTPException):
    """Indicate that there was an error with the incomplete HTTP request."""

    def __init__(self, original_exception, request_args, request_kwargs):
        """
        Args:
            original_exception: The original exception that occurred.
            request_args: The arguments to the request function.
            request_kwargs: The keyword arguments to the request function.
        """
        self.original_exception = original_exception
        self.request_args = request_args
        self.request_kwargs = request_kwargs
        super(RequestException, self).__init__(
            f"Error with request {original_exception}"
        )


class ResponseException(HTTPException):
    """Indicate that there was an error with the completed HTTP request."""

    def __init__(self, response: "Response"):
        self.response = response
        super(ResponseException, self).__init__(
            f"Received {response.status_code} HTTP response"
        )


class BadRequest(ResponseException):
    """Indicate a malformed request."""


class BadJSON(ResponseException):
    """Indicate the response did not contain valid JSON."""


class Unauthorized(ResponseException):
    """Indicate invalid credentials for the target resource."""


class Forbidden(ResponseException):
    """Indicate invalid permissions to complete the requested action."""


class NotFound(ResponseException):
    """Indicate that the requested resource was not found."""


class Conflict(ResponseException):
    """Indicate a conflicting change in the target resource."""


class Gone(ResponseException):
    """Indicate acess to the target resource is no longer available."""


class Unprocessable(ResponseException):
    """Indicate the request was not processable."""


class ServerError(ResponseException):
    """Indicate issues on the server end"""



"""Provides the User class."""
from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING, Optional, Dict, Any, Union

from .base import DriveBase
from .plan import Plan

if TYPE_CHECKING:
    from .. import TekDrive


@dataclass
class PartialUser:
    """
    Represents a simple User which provides a subset of a full User's attributes.

    Attributes:
        id (str): Unique ID for the user
        username (str): Username for the user
    """

    id: str
    username: str


class DriveUser(DriveBase):
    """
    Represents a TekDrive full user.

    Attributes:
        id (str): Unique user ID.
        username (str): Username (email) of the user.
        account_id (str): TekDrive account ID for the user.
        created_at (datetime): When the user account was created.
        updated_at (datetime): When the user account was updated.
        plan (:ref:`plan`): Plan details for the user.
    """

    STR_FIELD = "id"

    def __init__(
        self,
        tekdrive: "TekDrive",
        _data: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(tekdrive, _data=_data)

    def __setattr__(
        self,
        attribute: str,
        value: Union[str, int, Dict[str, Any]],
    ):
        if attribute == "plan":
            value = Plan.from_data(self._tekdrive, value)
        elif attribute in ("created_at", "updated_at", "shared_at"):
            if value is not None:
                value = datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ")
        super().__setattr__(attribute, value)



"""Provide the TekDrive client"""
import logging
from typing import TYPE_CHECKING, IO, Any, Dict, Optional, Union

from .authorizer import AccessKeyAuthorizer
from .session import create_session

from . import models
from .settings import TIMEOUT, BASE_URL
from .exceptions import (
    ClientException,
    ResponseException,
    TekDriveAPIException,
)
from .models.parser import Parser
from .utils.casing import to_snake_case

if TYPE_CHECKING:
    from .routing import Route

logging.getLogger("tekdrive").addHandler(logging.NullHandler())


class TekDrive:
    def __init__(
        self, access_key: str, base_url: str = BASE_URL, debug_mode: bool = False
    ):
        """
        Initialize a TekDrive instance.

        Args:
            access_key: Previously generated TekDrive access key.
            base_url: Base url for the TekDrive API.
            debug_mode: Should enable debug logging?
        """
        if not access_key:
            raise ClientException("Missing required attribute 'access_key'.")

        if debug_mode:
            logging.basicConfig(level=logging.DEBUG)

        # create authorizer and session
        self._authorizer = AccessKeyAuthorizer(access_key=access_key)
        self._session = create_session(authorizer=self._authorizer, base_url=base_url)

        # prepare parser
        self._parser = Parser(self, self._create_model_map())

        # models and helpers
        self.file = models.FileHelper(self, None)
        self.folder = models.FolderHelper(self, None)
        self.search = models.Search(self)
        self.trash = models.Trashcan(self)
        self.tree = models.Tree(self)
        self.user = models.User(self)

    def __enter__(self):
        """Context manager enter"""
        return self

    def __exit__(self, *_args):
        """Context manager exit"""
        pass

    def _create_model_map(self):
        model_map = {
            "Artifact": models.Artifact,
            "ArtifactsList": models.ArtifactsList,
            "File": models.File,
            "Folder": models.Folder,
            "Member": models.Member,
            "MembersList": models.MembersList,
            "PaginatedList": models.PaginatedList,
            "TrashPaginatedList": models.TrashPaginatedList,
            "Trash": models.Trash,
            "DriveUser": models.DriveUser,
        }
        return model_map

    def _request(
        self,
        route: "Route",
        params: Optional[Union[str, Dict[str, Union[str, int]]]] = None,
        data: Optional[Union[Dict[str, Union[str, Any]], bytes, IO, str]] = None,
        headers: Optional[Dict[str, Union[str, Any]]] = None,
        files: Optional[Dict[str, IO]] = None,
        json=None,
    ):
        if data and json:
            raise ClientException("Only supply one of: 'json', 'data'.")
        try:
            return self._session.request(
                route=route,
                data=data,
                files=files,
                params=params,
                headers=headers,
                timeout=TIMEOUT,
                json=json,
            )
        except ResponseException as exception:
            try:
                error_info = exception.response.json()
            except ValueError:
                raise Exception("Unexpected ResponseException") from exception

            api_error = self._parser.parse_error(
                error_info, headers=exception.response.headers
            )
            if api_error:
                # expected error format from API with known error code
                raise api_error from exception
            else:
                # raise generic api exception
                raise TekDriveAPIException(
                    to_snake_case(error_info), headers=exception.response.headers
                ) from exception

    def request(
        self,
        route: "Route",
        params: Optional[Union[str, Dict[str, Union[str, int]]]] = None,
        data: Optional[Union[Dict[str, Union[str, Any]], bytes, IO, str]] = None,
        headers: Optional[Dict[str, Union[str, Any]]] = None,
        files: Optional[Dict[str, IO]] = None,
        json=None,
        should_parse: bool = True,
    ) -> Any:
        """
        Return JSON data returned from a request using the provided route.

        Args:
            method: The HTTP method (e.g., GET, POST, PUT, DELETE).
            path: The path to fetch.
            params: The query parameters to add to the request.
            data: Dictionary, bytes, or file-like object to send in the body of the
                request.
            files: Dictionary, filename to file (like) object mapping
            json: JSON-serializable object to send in the body of the request with a
                Content-Type header of application/json. If ``json`` is provided,
                ``data`` should not be.
            should_parse: Should the response be parsed into a TekDrive model?
        """
        data = self._request(
            route=route,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
        )
        if should_parse:
            return self._parser.parse(data)
        else:
            return data



import logging
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union

from ..enums import ErrorCode, ObjectType
from ..exceptions import TekDriveAPIException, ERROR_CODE_TO_API_EXCEPTION_MAPPING
from ..utils.casing import to_snake_case

if TYPE_CHECKING:
    from .. import TekDrive

log = logging.getLogger(__name__)


class Parser:
    @classmethod
    def parse_error(
        cls, data: Union[List[Any], Dict[str, Dict[str, str]]], *, headers
    ) -> Optional[TekDriveAPIException]:
        """Convert JSON response into an API error"""
        data = to_snake_case(data)
        error_code = data.get("error_code")
        if error_code is None:
            # doesnt match expected error format from API
            return None

        try:
            error_code = ErrorCode(error_code)
        except ValueError:
            return TekDriveAPIException(data, headers=headers)

        if error_code in ERROR_CODE_TO_API_EXCEPTION_MAPPING:
            return ERROR_CODE_TO_API_EXCEPTION_MAPPING[error_code](
                to_snake_case(data), headers=headers
            )

        return TekDriveAPIException(data, headers=headers)

    def __init__(self, tekdrive: "TekDrive", models: Optional[Dict[str, Any]] = None):
        self._tekdrive = tekdrive
        self.models = {} if models is None else models

    def _is_artifact(self, data: dict) -> bool:
        return data.get("type") == ObjectType.ARTIFACT.value

    def _is_artifacts_list(self, data: dict) -> bool:
        return "artifacts" in data

    def _is_file(self, data: dict) -> bool:
        return data.get("type") == ObjectType.FILE.value

    def _is_file_with_upload_url(self, data: dict) -> bool:
        return {"file", "upload_url"}.issubset(data)

    def _is_folder(self, data: dict) -> bool:
        return data.get("type") == ObjectType.FOLDER.value

    def _is_members_list(self, data: dict) -> bool:
        return "members" in data

    def _is_member(self, data: dict) -> bool:
        return {"id", "username", "permissions"}.issubset(data)

    def _is_paginated_list(self, data: dict) -> bool:
        return "meta" in data

    def _is_trash(self, data: dict) -> bool:
        return "trasher" in data

    def _is_drive_user(self, data: dict) -> bool:
        return {"account_id", "owner_type", "plan"}.issubset(data)

    def _is_tree(self, data: dict) -> bool:
        return "tree" in data

    def _parse_dict(self, data: dict):
        """Create model from dict."""
        data = to_snake_case(data)
        if self._is_file(data):
            model = self.models["File"]
        elif self._is_folder(data):
            model = self.models["Folder"]
        elif self._is_artifacts_list(data):
            model = self.models["ArtifactsList"]
        elif self._is_artifact(data):
            model = self.models["Artifact"]
        elif self._is_members_list(data):
            model = self.models["MembersList"]
        elif self._is_member(data):
            model = self.models["Member"]
        elif self._is_file_with_upload_url(data):
            model = self.models["File"]
            file = data["file"]
            file["_upload_url"] = data["upload_url"]
            data = file
        elif self._is_paginated_list(data):
            if "trash" in data:
                model = self.models["TrashPaginatedList"]
            else:
                # generic paginated list
                model = self.models["PaginatedList"]
        elif self._is_trash(data):
            data["id"] = f"trash{data['item']['id']}"
            model = self.models["Trash"]
        elif self._is_drive_user(data):
            model = self.models["DriveUser"]
        elif self._is_tree(data):
            tree_folder = data["tree"]
            data = tree_folder
            model = self.models["Folder"]
        else:
            log.debug(f"Parsing found unknown model for: {data}")
            return data
        log.debug(f"Parsing using model: {model}")
        return model.parse(data, self._tekdrive)

    def parse(
        self, data: Optional[Union[Dict[str, Any], List[Any]]]
    ) -> Optional[Union[Dict[str, Any], List[Any]]]:
        log.debug(f"Parse data: {data}")
        if data is None:
            # HTTP 204 No Content
            return None

        if isinstance(data, list):
            return [self.parse(item) for item in data]

        if isinstance(data, dict):
            return self._parse_dict(data)

        return data



from .drive.file import File  # noqa
from .drive.folder import Folder  # noqa
from .drive.artifact import Artifact, ArtifactsList  # noqa
from .drive.member import Member, MembersList  # noqa
from .drive.plan import Plan  # noqa
from .drive.trash import Trash  # noqa
from .drive.user import PartialUser, DriveUser  # noqa
from .helpers import FileHelper, FolderHelper  # noqa
from .paginator import PaginatedList, TrashPaginatedList  # noqa
from .parser import Parser  # noqa
from .permissions import Permissions  # noqa
from .search import Search  # noqa
from .trashcan import Trashcan  # noqa
from .tree import Tree  # noqa
from .usage import Usage  # noqa
from .user import User  # noqa



"""Provides the Member class."""
from typing import TYPE_CHECKING, Optional, Dict, Any, Union

from .base import DriveBase
from ..base import BaseList
from ..permissions import Permissions

if TYPE_CHECKING:
    from .. import TekDrive


class Member(DriveBase):
    """
    Represents a user who has access to a TekDrive file or folder along
    with their permissions for that object.

    Attributes:
        id (str): Unique user ID
        username (str): Username of the member
        permissions (:ref:`permissions`): Member permissions
    """

    STR_FIELD = "id"

    @classmethod
    def from_data(cls, tekdrive, data):
        return cls(tekdrive, data)

    def __init__(
        self,
        tekdrive: "TekDrive",
        _data: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(tekdrive, _data=_data)

    def __setattr__(
        self,
        attribute: str,
        value: Union[str, int, Dict[str, Any]],
    ):
        if attribute == "permissions":
            value = Permissions(**value)
        super().__setattr__(attribute, value)


class MembersList(BaseList):
    """List of members"""

    _parent = None

    LIST_ATTRIBUTE = "members"



"""Provides the Usage dataclass."""
from dataclasses import dataclass


@dataclass
class Usage:
    """
    Represents TekDrive usage details and stats.
    """

    total_bytes_owned: str
    files_owned_count: int
    total_bytes_owned_in_trash: str
    files_owned_in_trash_count: int
    total_bytes_created: str
    files_created_count: int
    storage_size_limit: str

    @property
    def storage_limit(self) -> int:
        """
        The total limit in bytes of the user's storage space.
        """
        return int(self.storage_size_limit)

    @property
    def storage_freeable(self) -> int:
        """
        The total number of bytes counting against the user's storage limit that
        are in the trash and can be freed by emptying the trash.
        """
        return int(self.total_bytes_owned_in_trash)

    @property
    def storage_used(self) -> int:
        """
        The current number of bytes counting against the user's storage limit.
        """
        return int(self.total_bytes_owned)

    @property
    def storage_remaining(self) -> int:
        """
        The total number of bytes of storage available.
        """
        return self.storage_limit - self.storage_used



"""Provides the Search class."""
from typing import TYPE_CHECKING, Iterator, List, Optional, Union

from ..routing import Route, ENDPOINTS
from .base import TekDriveBase
from .paginator import PaginatedListGenerator
from . import File, Folder
from ..exceptions import ClientException
from ..utils.casing import to_camel_case

if TYPE_CHECKING:
    from .. import TekDrive


class Search(TekDriveBase):
    """
    Provides various methods to help with searching for TekDrive files and folders.
    """

    def __init__(self, tekdrive: "TekDrive"):
        super().__init__(tekdrive=tekdrive, _data=None)

    def files(
        self,
        *,
        name: Optional[str] = None,
        limit: Optional[int] = 100,
        folder_id: [str] = None,
        silo: Optional[str] = None,
        depth: Optional[int] = 1,
        file_type: Optional[List[str]] = None,
        upload_state: Optional[List[str]] = None,
        order_by: Optional[List[str]] = None,
        include_trashed: bool = False,
    ) -> Iterator[File]:
        """
        Convenience method for files search.

        Args:
            limit: Total limit for returned results.
            name: File name to match on. Case insensitive.
            folder_id: Unique ID of folder to perform search within.
            silo: Name of the silo to perform the search within. Values: ``"SHARES"`` or ``"PERSONAL"``.
            depth: How many levels deep to perform search when specifying a ``folder_id`` or ``silo``.
            file_type: Limit results to files matching the given file type(s).
            upload_state: Limit results to files in the given upload state(s).

        Examples:
            Get up to 50 WFM files::

                results = td.search.files(file_type="WFM", limit=50)

        Returns:
            Iterator [ :ref:`file` ]

        """
        return self.query(
            limit=limit,
            include_files=True,
            include_folders=False,
            folder_id=folder_id,
            silo=silo,
            depth=depth,
            file_type=file_type,
            upload_state=upload_state,
            name=name,
            order_by=order_by,
            include_trashed=include_trashed,
        )

    def folders(
        self,
        *,
        name: Optional[str] = None,
        limit: Optional[int] = 100,
        folder_id: [str] = None,
        silo: Optional[str] = None,
        depth: Optional[int] = 1,
        order_by: Optional[List[str]] = None,
        include_trashed: bool = False,
    ) -> Iterator[Folder]:
        """
        Convenience method for folders search.

        Args:
            limit: Total limit for returned results.
            silo: Name of the silo to perform the search within. Values: ``"SHARES"`` or ``"PERSONAL"``.
            name: Folder name to match on. Case insensitive.
            folder_id: Unique ID of folder to perform search within.
            depth: How many levels deep to perform search when specifying a ``folder_id`` or ``silo``.

        Examples:
            Get up to 10 folders with a name like ``"team_"``::

                results = td.search.folders(name="team_", limit=10)

        Returns:
            Iterator [ :ref:`folder` ]
        """
        return self.query(
            limit=limit,
            include_files=False,
            include_folders=True,
            folder_id=folder_id,
            silo=silo,
            depth=depth,
            name=name,
            order_by=order_by,
            include_trashed=include_trashed,
        )

    def query(
        self,
        *,
        name: Optional[str] = None,
        limit: Optional[int] = 100,
        folder_id: [str] = None,
        silo: Optional[str] = None,
        depth: Optional[int] = 1,
        file_type: Optional[List[str]] = None,
        include_files: bool = True,
        include_folders: bool = True,
        upload_state: Optional[List[str]] = None,
        order_by: Optional[List[str]] = None,
        include_trashed: bool = False,
    ) -> Iterator[Union[File, Folder]]:
        """
        Execute search for files and/or folders matching the provided criteria. A global search will
        be performed by default unless a ``folder_id`` or ``silo`` is provided.

        Args:
            limit: Total limit for returned results.
            name: File/Folder name to match on. Case insensitive.
            folder_id: Unique ID of folder to perform search within.
            depth: How many levels deep to perform search when specifying a ``folder_id`` or ``silo``.
            silo: Name of the silo to perform the search within. Values: ``"SHARES"`` or ``"PERSONAL"``.
            file_type: Limit results to files matching the given file type(s).
            include_files: Include files in the search results? Default: ``True``.
            include_folders: Include folders in the search results? Default: ``True``.
            upload_state: Limit results to files in the given upload state(s).
            include_trashed: Include files and folders in the trashcan.

        Examples:
            Get files with name like ``"project1"``::

                results = td.search.query(name="project1", include_folders=False)

        Returns:
            Iterator [ Union [ :ref:`file` , :ref:`folder` ] ]
        """
        if all(param is None for param in [name, file_type]):
            raise ClientException("Must supply `name`, `file_type`, or `upload_state`.")

        route = Route("GET", ENDPOINTS["search"])

        search_types = []
        if include_files is True:
            search_types.append("FILE")
        if include_folders is True:
            search_types.append("FOLDER")

        params = to_camel_case(
            dict(
                folder_id=folder_id,
                name=name,
                silo=silo,
                depth=depth,
                file_type=file_type,
                type=",".join(search_types),
                upload_state=upload_state,
                order_by=order_by,
                include_trashed=include_trashed,
            )
        )
        return PaginatedListGenerator(self._tekdrive, route, limit=limit, params=params)



"""Provides the File class."""
import os
import requests
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, IO, Optional, Union

from ...routing import Route, ENDPOINTS
from ...exceptions import ClientException, TekDriveStorageException
from ...utils.casing import to_snake_case, to_camel_case
from .base import DriveBase, Downloadable
from .artifact import Artifact, ArtifactsList
from .member import Member, MembersList
from .user import PartialUser
from ...enums import ObjectType
from ..permissions import Permissions

if TYPE_CHECKING:
    from .. import TekDrive


class File(DriveBase, Downloadable):
    """
    A class representing a TekDrive file.

    Examples:
        Load a file by id::

            file_id = "9606be66-8af0-42fc-b199-7c0ca7e30d73"
            file = td.file(file_id)

    Attributes:
        bytes (str): File size in bytes.
        created_at (datetime): When the file was created.
        creator (:ref:`partial_user`): File creator.
        file_type (str): File type such as ``"JPG"`` or ``"WFM"``.
        id (str): Unique ID of the file.
        name (str): Name of the file.
        owner (:ref:`partial_user`): File owner.
        parent_folder_id (str): Unique ID of the file's parent folder.
        shared_at (datetime, optional): When the file was shared with the
            requesting user. Will be ``None`` if the user has direct access.
        type (str): Type of TekDrive object - will always be ``"FILE"``.
        permissions (:ref:`permissions`): File permissions for the requesting user.
        updated_at (datetime, optional): When the file was last updated.
    """

    STR_FIELD = "id"

    def __init__(
        self,
        tekdrive: "TekDrive",
        id: Optional[str] = None,
        _data: Optional[Dict[str, Any]] = None,
    ):
        fetched = False
        if id:
            self.id = id
        elif _data:
            fetched = True
        else:
            raise ClientException("Must supply `id` or `_data`.")

        self._upload_url = None

        super().__init__(tekdrive, _data=_data, _fetched=fetched)

    def __setattr__(
        self,
        attribute: str,
        value: Union[str, int, Dict[str, Any]],
    ):
        if attribute == "owner" or attribute == "creator":
            value = PartialUser(**value)
        elif attribute in ("created_at", "updated_at", "shared_at"):
            if value is not None:
                value = datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ")
        elif attribute == "permissions":
            value = Permissions(**value)
        elif attribute == "type":
            value = ObjectType(value)
        super().__setattr__(attribute, value)

    def _fetch_data(self):
        route = Route("GET", ENDPOINTS["file_details"], file_id=self.id)
        return self._tekdrive.request(route, should_parse=False)

    def _fetch(self):
        data = self._fetch_data()
        other = type(self)(self._tekdrive, _data=to_snake_case(data))
        self.__dict__.update(other.__dict__)
        self._fetched = True

    def _fetch_upload_url(self):
        route = Route("GET", ENDPOINTS["file_upload"], file_id=self.id)
        upload_details = self._tekdrive.request(route)
        return upload_details["upload_url"]

    def _fetch_download_url(self):
        route = Route("GET", ENDPOINTS["file_download"], file_id=self.id)
        download_details = self._tekdrive.request(route)
        return download_details["download_url"]

    def _update_details(self, data):
        route = Route("PUT", ENDPOINTS["file_details"], file_id=self.id)
        return self._tekdrive.request(route, json=data)

    def _upload_to_storage(self, file: IO):
        # we may already have upload url from file creation
        if self._upload_url is None:
            self._upload_url = self._fetch_upload_url()

        try:
            r = requests.put(
                self._upload_url,
                data=file,
                headers={
                    "Content-Type": "application/octet-stream",
                },
            )
            r.raise_for_status()
        except requests.exceptions.HTTPError as exception:
            raise TekDriveStorageException("Upload failed") from exception

    @staticmethod
    def _create(
        _tekdrive,
        path_or_readable: Union[str, IO] = None,
        name: str = None,
        parent_folder_id: str = None,
    ) -> "File":
        if path_or_readable and name is None:
            # get name from path or readable
            if isinstance(path_or_readable, str):
                name = os.path.basename(path_or_readable)
            else:
                name = os.path.basename(path_or_readable.name)

        data = dict(name=name, parentFolderId=parent_folder_id)
        route = Route("POST", ENDPOINTS["file_create"])
        new_file = _tekdrive.request(route, json=data)

        if path_or_readable:
            new_file.upload(path_or_readable)

        return new_file

    def artifacts(self, flat: bool = False) -> ArtifactsList:
        """
        Get a list of file artifacts.

        Args:
            flat: Return artifacts in flat list with no child nesting?

        Examples:
            Iterate over all file artifacts::

                for artifact in file.artifacts():
                    print(artifact.name)

        Returns:
            List [ :ref:`artifact` ]
        """
        params = to_camel_case(dict(flat=flat))

        route = Route("GET", ENDPOINTS["file_artifacts"], file_id=self.id)
        artifacts = self._tekdrive.request(route, params=params)
        artifacts._parent = self
        return artifacts

    def artifact(self, artifact_id: str, depth: int = 1) -> Artifact:
        """
        Get a file artifact by ID.

        Args:
            artifact_id: Unique ID for the artifact
            depth: How many nested levels of child artifacts to return.

        Examples:
            Load artifact with children up to 3 levels deep::

                artifact_id = "017820c4-03ba-4e9d-be2f-e0ba346ddd9b"
                artifact = file.artifact(artifact_id, depth=3)

        Returns:
            :ref:`artifact`
        """
        params = to_camel_case(dict(depth=depth))

        route = Route("GET", ENDPOINTS["file_artifact"], file_id=self.id, artifact_id=artifact_id)
        artifact = self._tekdrive.request(route, params=params)
        return artifact

    def members(self) -> MembersList:
        """
        Get a list of file members.

        Examples:
            Iterate over all file members::

                for member in file.members():
                    print(member.username)

        Returns:
            List [ :ref:`member` ]
        """
        route = Route("GET", ENDPOINTS["file_members"], file_id=self.id)
        members = self._tekdrive.request(route)
        members._parent = self
        return members

    def restore(self) -> None:
        """
        Restore the file from user's trashcan.

        Examples:
            Restore - remove from trash::

                file.restore()
        """
        route = Route("POST", ENDPOINTS["file_restore"], file_id=self.id)
        self._tekdrive.request(route)

    def delete(self, hard_delete: bool = False) -> None:
        """
        Delete the file, by default it will be placed in the user's trashcan.

        Args:
            hard_delete: Permanently delete the file?

        Examples:
            Delete - placing in trash::

                file.delete()

            Delete - immediately delete::

                file.delete(hard_delete=True)
        """
        params = to_camel_case(dict(hard_delete=hard_delete))

        route = Route("DELETE", ENDPOINTS["file_delete"], file_id=self.id)
        self._tekdrive.request(route, params=params)

    def upload(self, path_or_readable: Union[str, IO]) -> None:
        """
        Upload file contents. This will overwrite existing content, if any.

        Args:
            path_or_readable: Path to a local file or a readable stream
                representing the contents to upload.

        Raises:
            ClientException: If invalid file path is given.

        Examples:
            Upload using path::

                here = os.path.dirname(__file__)
                contents_path = os.path.join(here, "test_file_overwrite.txt")
                file.upload(contents_path)

            Upload using readable stream::

                with open("./test_file.txt", "rb") as f:
                    new_file.upload(f)
        """
        # TODO: multipart upload support
        if isinstance(path_or_readable, str):
            file_path = path_or_readable

            if not os.path.exists(file_path):
                raise ClientException(f"File '{file_path}' does not exist.")

            with open(file_path, "rb") as file:
                self._upload_to_storage(file)
        else:
            readable = path_or_readable
            self._upload_to_storage(readable)

    def move(self, parent_folder_id: str) -> None:
        """
        Move file to a different folder.

        Args:
            parent_folder_id: Unique ID of folder to move the file into.

        Examples:
            Move by ID::

                folder_id = "3b525331-9da7-4e8d-b045-4acdba8d9dc7"
                file.move(folder_id)

            Move to newly created folder::

                folder = td.folder.create("FolderA")
                file.move(folder.id)
        """
        data = dict(parentFolderId=parent_folder_id)
        self._update_details(data)
        self.parent_folder_id = parent_folder_id

    def save(self) -> None:
        """
        Save any changes to file meta. Supported attributes: ``name``.

        Examples:
            Rename a file::

                file.name = "my_new_name"
                file.save()
        """
        data = dict(name=self.name)
        self._update_details(data)

    def add_member(
        self, username: str = None, user_id: str = None, edit_access: bool = False
    ) -> Member:
        """
        Share the file with an existing or new user.

        Args:
            username: The username (email) of the sharee.
            user_id: The user ID of the sharee.
            edit_access: Give member edit access?

        Examples:
            Share with read only permissions (default)::

                file_member = file.add_member(username="read_only@example.com")

            Share with edit permissions::

                file_member = file.add_member(user_id="354bcafb-6c54-4a1f-9b94-a76f38b548e5", edit_access=True)

        Returns:
            :ref:`member`
        """
        data = dict(permissions=dict(read=True, edit=edit_access))
        if user_id:
            data["id"] = user_id
        elif username:
            data["username"] = username
        else:
            raise ClientException("Must supply `username` or `user_id`.")

        route = Route("POST", ENDPOINTS["file_members"], file_id=self.id)
        return self._tekdrive.request(route, json=data)

    def remove_member(self, user_id: str) -> None:
        """
        Revoke access for a current file member.

        Args:
            user_id: The user ID of the member.

        Examples:
            Remove member::

                file.remove_member(user_id="354bcafb-6c54-4a1f-9b94-a76f38b548e5")

        """
        route = Route(
            "DELETE", ENDPOINTS["file_member"], file_id=self.id, member_id=user_id
        )
        self._tekdrive.request(route)

    def modify_member(self, user_id: str, edit_access: bool) -> Member:
        """
        Modify an existing file member.

        Args:
            user_id: The user ID of the member.
            edit_access: Give member edit access?

        Examples:
            Grant edit access::

                updated_file_member = file.modify_member(user_id="354bcafb-6c54-4a1f-9b94-a76f38b548e5", edit_access=True)

            Revoke edit access::

                updated_file_member = file.modify_member(user_id="354bcafb-6c54-4a1f-9b94-a76f38b548e5", edit_access=False)

        Returns:
            :ref:`member`

        """
        route = Route(
            "PUT", ENDPOINTS["file_member"], file_id=self.id, member_id=user_id
        )
        data = dict(permissions=dict(read=True, edit=edit_access))
        return self._tekdrive.request(route, json=data)



import re
from collections.abc import Mapping
from typing import Union, Dict, List

SPLIT_RE = re.compile(r"([\-_\s]*[A-Z0-9]+[^A-Z\-_\s]+[\-_\s]*)")
ACRONYM_RE = re.compile(r"([A-Z]+)(?=[A-Z][a-z])")
UNDERSCORE_RE = re.compile(r"([^\-_\s])[\-_\s]+([^\-_\s])")


def to_camel_case(str_or_iter):
    """Convert a string, dict, or list of dicts to camel case"""
    if isinstance(str_or_iter, (list, Mapping)):
        return _process_keys(str_or_iter, to_camel_case)

    s = str(str_or_iter)
    if s.isnumeric():
        return str_or_iter

    if s.isupper():
        return str_or_iter

    return "".join(
        [
            s[0].lower() if not s[:2].isupper() else s[0],
            UNDERSCORE_RE.sub(lambda m: m.group(1) + m.group(2).upper(), s[1:]),
        ]
    )


def to_snake_case(str_or_iter: Union[str, Dict, List]):
    """Convert a string, dict, or list of dicts to snake case"""
    if isinstance(str_or_iter, (list, Mapping)):
        return _process_keys(str_or_iter, to_snake_case)

    s = str(str_or_iter)
    if s.isnumeric():
        return str_or_iter

    if s.isupper():
        return str_or_iter

    return _break_words(_fix_abbrevations(s)).lower()


def _break_words(string: str, separator: str = "_", split=SPLIT_RE.split):
    return separator.join(s for s in split(string) if s)


def _fix_abbrevations(string: str) -> str:
    """
    Fix acronyms, initialisms, and abbrevations, e.g. "ObjectID" should
    be "object_id" not "object_i_d"
    """
    return ACRONYM_RE.sub(lambda m: m.group(0).title(), string)


def _process_keys(str_or_iter, fn):
    if isinstance(str_or_iter, list):
        return [_process_keys(k, fn) for k in str_or_iter]
    elif isinstance(str_or_iter, Mapping):
        return {fn(k): _process_keys(v, fn) for k, v in str_or_iter.items()}
    else:
        return str_or_iter



"""Provides the User class."""
from typing import TYPE_CHECKING

from ..routing import Route, ENDPOINTS
from .base import TekDriveBase
from . import DriveUser
from .usage import Usage

if TYPE_CHECKING:
    from .. import TekDrive


class User(TekDriveBase):
    def __init__(self, tekdrive: "TekDrive"):
        super().__init__(tekdrive, _data=None)

    def me(self) -> DriveUser:
        """
        Get user details for the authenticated user.

        Examples:
            Fetch user details::

                me = td.user.me()

            Get the plan the user is on::

                plan = td.user.me().plan

        Returns: 
            :ref:`drive_user`

        """
        route = Route("GET", ENDPOINTS["user"])
        return self._tekdrive.request(route)

    def usage(self) -> Usage:
        """
        Get TekDrive usage details for the authenticated user.

        Examples:
            Fetch usage details::

                usage_info = td.user.usage()

        """
        route = Route("GET", ENDPOINTS["usage"])
        data = self._tekdrive.request(route)
        return Usage(
            total_bytes_owned=data["total_bytes_owned"],
            files_owned_count=data["files_owned_count"],
            total_bytes_owned_in_trash=data["total_bytes_owned_in_trash"],
            files_owned_in_trash_count=data["files_owned_in_trash_count"],
            total_bytes_created=data["total_bytes_created"],
            files_created_count=data["files_created_count"],
            storage_size_limit=data["storage_size_limit"],
        )



from copy import deepcopy
from .base import TekDriveBase
from typing import TYPE_CHECKING, Any, Dict, Iterator, Optional, Union

if TYPE_CHECKING:
    from .. import TekDrive
    from ..routing import Route


class PaginatedList(TekDriveBase):
    LIST_ATTRIBUTE = "results"
    META_ATTRIBUTE = "meta"

    def __len__(self) -> int:
        """Return the number of items in the full list."""
        return len(getattr(self, self.LIST_ATTRIBUTE))

    def __getitem__(self, index: int) -> Any:
        """Return the item at position index in the list."""
        return getattr(self, self.LIST_ATTRIBUTE)[index]

    def __setattr__(self, attribute: str, value: Any):
        """Parse the LIST_ATTRIBUTE attribute."""
        if attribute == self.LIST_ATTRIBUTE:
            value = self._tekdrive._parser.parse(value)
        super().__setattr__(attribute, value)

    @property
    def page(self) -> int:
        return getattr(self, self.META_ATTRIBUTE)["page"]

    @property
    def limit_per_page(self) -> int:
        return getattr(self, self.META_ATTRIBUTE)["limit"]


class TrashPaginatedList(PaginatedList):
    LIST_ATTRIBUTE = "trash"
    META_ATTRIBUTE = "meta"


class PaginatedListGenerator(TekDriveBase, Iterator):
    def __init__(
        self,
        tekdrive: "TekDrive",
        route: "Route",
        limit: int = 100,
        limit_per_page: int = 100,
        params: Optional[Dict[str, Union[str, int]]] = None,
    ):
        """
        Initialize a PaginatedListGenerator instance.

        Args:
            tekdrive: An instance of :class:`.TekDrive`.
            route: A Route for an API endpoint returning a paginated list.
            limit: Number of total results to fetch.
            params: A dictionary containing additional query string parameters to
                send with the request.
        """
        super().__init__(tekdrive, _data=None)
        self._exhausted = False
        self._list = None
        self._list_index = None
        self.limit = limit  # total results limit
        self.params = deepcopy(params) if params else {}
        self.params["limit"] = min(limit, limit_per_page)  # limit for a single page
        self.route = route
        self.yielded = 0

    def __iter__(self) -> Iterator[Any]:
        """Permit ListingGenerator to operate as an iterator."""
        return self

    def __next__(self) -> Any:
        """Permit ListingGenerator to operate as a generator."""
        if self.limit is not None and self.yielded >= self.limit:
            raise StopIteration()

        if self._list is None or self._list_index >= len(self._list):
            self._next_batch()

        self._list_index += 1
        self.yielded += 1
        return self._list[self._list_index - 1]

    def _next_batch(self):
        if self._exhausted:
            raise StopIteration()

        self._list = self._tekdrive.request(self.route, params=self.params)
        self._list_index = 0

        if not self._list:
            raise StopIteration()

        if len(self._list) == self._list.limit_per_page:
            # go to next page
            self.params["page"] = self._list.page + 1
        else:
            self._exhausted = True



from requests.exceptions import (
    ChunkedEncodingError,
    ConnectionError,
    Timeout,  # both ConnectTimeout and ReadTimeout errors
)
from requests.status_codes import codes as http_codes

from .exceptions import (
    BadRequest,
    Unauthorized,
    Forbidden,
    NotFound,
    Conflict,
    Gone,
    Unprocessable,
    ServerError,
)

NO_CONTENT = http_codes.NO_CONTENT
SUCCESS_STATUS_CODES = (http_codes.CREATED, http_codes.OK)
RETRY_STATUS_CODES = (
    http_codes.BAD_GATEWAY,
    http_codes.GATEWAY_TIMEOUT,
    http_codes.INTERNAL_SERVER_ERROR,
    http_codes.SERVICE_UNAVAILABLE,
    520,  # Web server is returning an unknown error a.k.a "catch-all"
    522,  # Connection timed out
)
RETRY_EXCEPTIONS = (ChunkedEncodingError, ConnectionError, Timeout)

STATUS_TO_EXCEPTION_MAPPING = {
    http_codes.BAD_GATEWAY: ServerError,
    http_codes.BAD_REQUEST: BadRequest,
    http_codes.CONFLICT: Conflict,
    http_codes.FORBIDDEN: Forbidden,
    http_codes.NOT_FOUND: NotFound,
    http_codes.GONE: Gone,
    http_codes.GATEWAY_TIMEOUT: ServerError,
    http_codes.INTERNAL_SERVER_ERROR: ServerError,
    http_codes.SERVICE_UNAVAILABLE: ServerError,
    http_codes.UNAUTHORIZED: Unauthorized,
    http_codes.UNPROCESSABLE_ENTITY: Unprocessable,
}
EXCEPTION_STATUS_CODES = [code for code in STATUS_TO_EXCEPTION_MAPPING.keys()]



"""Provide Session class."""

import logging
from time import sleep
from copy import deepcopy
from typing import TYPE_CHECKING, Dict, Optional, Tuple, Union
from urllib.parse import urljoin

from .authorizer import BaseAuthorizer
from .request_wrapper import RequestWrapper
from .retry import RetryPolicy, RateLimit
from .exceptions import (
    InvalidAuthorizer,
    RequestException,
    BadJSON,
)
from .status_codes import (
    EXCEPTION_STATUS_CODES,
    NO_CONTENT,
    RETRY_EXCEPTIONS,
    RETRY_STATUS_CODES,
    STATUS_TO_EXCEPTION_MAPPING,
    SUCCESS_STATUS_CODES,
)
from .settings import TIMEOUT, BASE_URL

if TYPE_CHECKING:
    from requests import Response
    from .routing import Route

log = logging.getLogger(__name__)


class Session(object):
    def __init__(self, authorizer: BaseAuthorizer, base_url: str = BASE_URL):
        if not isinstance(authorizer, BaseAuthorizer):
            raise InvalidAuthorizer(f"Invalid Authorizer: {authorizer}")

        self._authorizer = authorizer
        self._rate_limit = RateLimit()
        self._request_wrapper = RequestWrapper(base_url=base_url)

    def _log_request(self, method, url, params, data, json) -> None:
        log.debug(
            f"Request: {method} {url}, data: {data}, json: {json}, params: {params}"
        )

    def __enter__(self):
        """Context manager enter"""
        return self

    def __exit__(self, *_args):
        """Context manager exit"""
        self.close()

    def _try_request(
        self,
        *,
        method,
        url,
        data,
        files,
        json,
        params,
        headers,
        retry_policy,
        timeout,
    ) -> Tuple[Optional["Response"], Optional[Exception]]:
        if not headers:
            headers = self._authorizer._get_auth_header()

        seconds_to_sleep = self._rate_limit.seconds_to_sleep()
        if seconds_to_sleep:
            log.debug(f"Sleeping for {seconds_to_sleep} seconds (rate limited)")
            sleep(seconds_to_sleep)
        try:
            response = self._request_wrapper.call(
                method,
                url,
                data=data,
                files=files,
                json=json,
                params=params,
                headers=headers,
                timeout=timeout,
            )
            log.debug(f"Response status: {response.status_code}")

            # update the rate limit state from response headers
            self._rate_limit.update_from_headers(response.headers)

            return response, None
        except RequestException as exception:
            if not retry_policy.retries_remaining or not isinstance(
                exception.original_exception, RETRY_EXCEPTIONS
            ):
                raise
            return None, exception.original_exception

    def _request(
        self,
        *,
        method,
        url,
        data,
        files,
        json,
        params,
        headers,
        timeout,
    ):
        retry_policy = RetryPolicy()
        self._log_request(method, url, params, data, json)

        while retry_policy.retries_remaining:
            response, exc = self._try_request(
                method=method,
                url=url,
                data=data,
                files=files,
                json=json,
                params=params,
                headers=headers,
                retry_policy=retry_policy,
                timeout=timeout,
            )
            if response is None or response.status_code not in RETRY_STATUS_CODES:
                break

            retry_policy.decrement_retries()
            sleep_seconds = retry_policy.seconds_to_sleep()
            if sleep_seconds > 0:
                sleep(sleep_seconds)

        status_code = repr(exc) if exc else response.status_code
        if status_code in EXCEPTION_STATUS_CODES:
            raise STATUS_TO_EXCEPTION_MAPPING[response.status_code](response)
        elif status_code == NO_CONTENT:
            return
        elif status_code not in SUCCESS_STATUS_CODES:
            raise Exception(f"Unknown status code: {status_code}")

        try:
            return response.json()
        except ValueError:
            return BadJSON(response)

    def close(self):
        self._request_wrapper.close()

    def safe_copy_dict(self, d: dict, sort: bool = False) -> dict:
        if isinstance(d, dict):
            d = deepcopy(d)
            if sort:
                return sorted(d.items())
        return d

    def request(
        self,
        route: "Route",
        data: dict = None,
        files: dict = None,
        json: object = None,
        params: Optional[Union[str, Dict[str, Union[str, int]]]] = None,
        headers: dict = None,
        timeout: float = TIMEOUT,
    ):
        """
        Return the json content from the resource at ``path``.

        Args:
            method: The request method such as GET, POST, PATCH, etc.
            path: The path of the request to be combined with the ``base_url``
                of the Requestor.
            data: Dictionary, bytes, or file-like object to send in the body
                of the request.
            files: Dictionary, mapping ``filename`` to file-like object.
            json: Object to be serialized to JSON in the body of the
                request.
            params: The query parameters to send with the request.
            headers: Overwrite all headers for the request.
        """
        params = deepcopy(params) or {}
        data = self.safe_copy_dict(data, sort=True)
        headers = self.safe_copy_dict(headers, sort=True)
        json = self.safe_copy_dict(json)

        return self._request(
            method=route.method,
            url=urljoin(self._request_wrapper.base_url, route.path),
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            timeout=timeout,
        )


def create_session(*, authorizer: BaseAuthorizer = None, base_url: str) -> Session:
    return Session(authorizer, base_url=base_url)



from typing import IO, Union

from .base import TekDriveBase
from .drive.file import File
from .drive.folder import Folder
from ..exceptions import ClientException


class FileHelper(TekDriveBase):
    def __call__(self, id: str) -> File:
        """
        Return a lazy instance of File.

        Args:
            name: The name of the file.
        """
        return File(self._tekdrive, id=id)

    def create(
        self,
        path_or_readable: Union[str, IO] = None,
        name: str = None,
        parent_folder_id: str = None,
    ) -> File:
        """
        Create a new file.
        """
        if path_or_readable is None and name is None:
            raise ClientException("Must supply 'path_or_readable' or 'name'.")

        return File._create(
            _tekdrive=self._tekdrive,
            path_or_readable=path_or_readable,
            name=name,
            parent_folder_id=parent_folder_id,
        )


class FolderHelper(TekDriveBase):
    def __call__(self, id: str) -> Folder:
        """
        Return a lazy instance of Folder.

        Args:
            name: The name of the folder.
        """
        return Folder(self._tekdrive, id=id)

    def create(
        self,
        name: str = None,
        parent_folder_id: str = None,
    ) -> Folder:
        """
        Create a new folder.
        """
        return Folder._create(
            _tekdrive=self._tekdrive,
            name=name,
            parent_folder_id=parent_folder_id,
        )



"""Provide retry policy"""
import logging
import time
from dataclasses import dataclass
from random import random

log = logging.getLogger(__name__)


@dataclass
class RetryPolicy:
    """
    Track retries for API requests
    """

    retries: int = 2
    max_backoff: int = 32

    def seconds_to_sleep(self):
        return min(((2 ** self.retries) + random()), self.max_backoff)

    @property
    def retries_remaining(self):
        return self.retries >= 1

    def decrement_retries(self):
        self.retries -= 1


@dataclass
class RateLimit:
    reset_timestamp = None
    next_request_timestamp = None
    remaining: int = None
    used: int = None
    limit: int = None

    def update_from_headers(self, headers: dict):
        """
        Update rate limit fields
        """
        log.debug(f"Update rate limit from headers: {headers}")
        if "x-ratelimit-remaining" not in headers:
            if self.remaining is not None:
                self.remaining -= 1
                self.used += 1
            return

        now = time.time()

        self.reset_timestamp = int(headers["x-ratelimit-reset"])
        self.limit = int(headers["x-ratelimit-limit"])
        self.remaining = int(headers["x-ratelimit-remaining"])
        self.used = self.limit - self.remaining

        if self.remaining <= 0:
            self.next_request_timestamp = self.reset_timestamp
            return

        self.next_request_timestamp = now

    def seconds_to_sleep(self):
        """
        How many seconds to sleep to avoid rate limit throttling?
        """
        if self.next_request_timestamp is None:
            return
        sleep_seconds = self.next_request_timestamp - time.time()
        if sleep_seconds <= 0:
            return
        return sleep_seconds



"""Provides the Member class."""
from typing import TYPE_CHECKING, Optional, Dict, Any, Union

from .base import DriveBase
from .file import File
from .folder import Folder

if TYPE_CHECKING:
    from .. import TekDrive


class Trash(DriveBase):
    """
    Represents a file or folder that has been placed in the trashcan.

    Attributes:
        trasher (PartialUser): The user who placed the object in the trash.
        trashed_at (datetime): When the object was placed in the trash.
        trashed_directly (bool): Was the item placed directly in the trash?
        total_bytes (str): The total bytes of the trashed object. If the item is
            a folder, this will be the total sum of the folder contents.
        item_share_count (int): How many TekDrive users the item is shared with.
        item (:ref:`file` or :ref:`folder`): File or folder details.
    """

    STR_FIELD = "id"

    @classmethod
    def from_data(cls, tekdrive, data):
        return cls(tekdrive, data)

    def __init__(
        self,
        tekdrive: "TekDrive",
        _data: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(tekdrive, _data=_data)

    def __setattr__(
        self,
        attribute: str,
        value: Union[str, int, Dict[str, Any]],
    ):
        if attribute == "item" and value.get("type") == "FILE":
            value = File(self._tekdrive, value["id"], _data=value)
        elif attribute == "item" and value.get("type") == "FOLDER":
            value = Folder(self._tekdrive, value["id"], _data=value)
        super().__setattr__(attribute, value)



"""Provides AccessKeyAuthorizer and other future authorizer classes."""


class BaseAuthorizer(object):
    def __init__(self):
        pass

    def _set_auth_header(self):
        raise NotImplementedError()


class AccessKeyAuthorizer(BaseAuthorizer):
    def __init__(self, access_key):
        """
        Represent a single, personal-use authorization via access key.

        Args:
            access_key: The previously generated access key.
        """
        self._access_key = access_key

    def _get_auth_header(self):
        return {"X-IS-AK": self._access_key}



import re
from os import path

from setuptools import find_packages, setup

PACKAGE_NAME = "tekdrive"
here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md"), "r", encoding="utf-8") as readme:
    README = readme.read()

with open(path.join(here, PACKAGE_NAME, "settings.py"), encoding="utf-8") as settings:
    VERSION = re.search('__version__ = "([^"]+)"', settings.read()).group(1)

extras = {
    "lint": [
        "black ==20.8b1",
        "flake8 >=3.8.4",
    ],
    "test": [
        "pytest >=6.2.2",
        "vcrpy >=4.1.1",
    ],
    "release": [
        "twine ==3.4.2",
    ]
}
extras["dev"] = extras["lint"] + extras["test"]

setup(
    name=PACKAGE_NAME,
    author="Initial State",
    author_email="thomas@initialstate.com",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Utilities",
    ],
    description=("Package to interact with the TekDrive API."),
    install_requires=["requests >=2.25.0, <3.0"],
    extras_require=extras,
    keywords="tektronix tekdrive tekcloud",
    long_description=README,
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=["tests", "tests.*"]),
    version=VERSION,
)



"""Provides the Permissions dataclass."""
from dataclasses import dataclass


@dataclass
class Permissions:
    """
    Represents permissions to a TekDrive object.

    Attributes:
        creator (bool): Is the object creator?
        edit (bool): Has edit access?
        read (bool): Has read access?
        owner (bool): Is the object owner?
    """

    read: bool
    edit: bool
    owner: bool = None
    creator: bool = None
    public: bool = None  # currently unused



from urllib.parse import quote as _uriquote


class Route:
    def __init__(self, method, path_template, **params):
        self.method = method
        self.path_template = path_template
        if params:
            self.path = path_template.format(
                **{
                    k: _uriquote(v) if isinstance(v, str) else v
                    for k, v in params.items()
                }
            )
        else:
            self.path = path_template

        # major params:
        self.file_id = params.get("file_id")
        self.folder_id = params.get("folder_id")
        self.member_id = params.get("member_id")


ENDPOINTS = {
    "file_artifacts": "/file/{file_id}/artifacts",
    "file_artifact": "/file/{file_id}/artifacts/{artifact_id}",
    "file_artifact_download": "/file/{file_id}/artifacts/{artifact_id}/contents",
    "file_create": "/file",
    "file_details": "/file/{file_id}",
    "file_delete": "/file/{file_id}",
    "file_restore": "/file/{file_id}",
    "file_download": "/file/{file_id}/contents",
    "file_members": "/file/{file_id}/members",
    "file_member": "/file/{file_id}/members/{member_id}",
    "file_upload": "/file/{file_id}/uploadUrl",
    "folder_create": "/folder",
    "folder_details": "/folder/{folder_id}",
    "folder_delete": "/folder/{folder_id}",
    "folder_restore": "/folder/{folder_id}",
    "folder_members": "/folder/{folder_id}/members",
    "folder_member": "/folder/{folder_id}/members/{member_id}",
    "search": "/search",
    "trash": "/trash",
    "tree": "/tree",
    "usage": "/user/usage",
    "user": "/user",
}
